/*
 *                   jfjji,.           H A M M E R H E A D  S H A R K W A R E
 *                  ,#######EL:                  http://www.hhsw.de
 *                   .,j######Kj       
 *                    ;W########G:     This  code  is  free software; you can
 *                   j#####G ,D#L:     redistribute it and/or modify it under
 *             .,;;,G#####D.           the  terms  of  the GNU General Public
 *           jE##########G             License   as  published  by  the  Free
 *             .,W######D.             Software  Foundation; either version 2
 *             :L#####K##G             of  the  License,  or (at your option)
 *           .f#####G;.t##:            any later version.
 *         ;L###ELi.    ,D,            
 *       iEEGWj                        This  code  is distributed in the hope
 *      tW;                            that  it  will  be useful, but without
 *     jf                              any warranty; without even the implied
 *    .Kt                              warranty of merchantability or fitness
 *    i:                               for  a particular purpose. See the GNU
 *    ;                                General   Public   License   for  more
 *    .                                details (http://www.gnu.org).
 */
#include <string>

#include "WindowManager.hh"

using namespace std;
using namespace bbdock;

/**
 * Check if window is valid
 *
 * @param w - window id
 */
const bool WindowManager::isValid( Window w ) const
{
	XErrorHandler handler;
	XWindowAttributes attr;
	bool valid = false;

	handler = XSetErrorHandler( WindowManager::throwError );
	if( XGetWindowAttributes( display, w, &attr ) )
		valid = true;
	XSetErrorHandler( handler );

	return valid;
}

/**
 * Activate a window and switch to its workspace
 *
 * @param w - id of window to activate
 */
void WindowManager::activateWindow( Window w ) const
{
	WindowManager::Property <unsigned long> p;

	// switch to workspace if neccessary
	if( p.getProperty( display, w, XA_CARDINAL, "_NET_WM_DESKTOP" ) ||
		p.getProperty( display, w, XA_CARDINAL, "_WIN_WORKSPACE" ) )
	{
		unsigned long workspace = *p.getData();

		if( (p.getProperty( display, root, XA_CARDINAL, 
			"_NET_CURRENT_DESKTOP" ) ||
			p.getProperty( display, root, XA_CARDINAL, "_WIN_WORKSPACE" )) &&
			workspace != *p.getData() )
			sendClientMessage( root, "_NET_CURRENT_DESKTOP", workspace );
	}

	sendClientMessage( w, "_NET_ACTIVE_WINDOW" );
	XMapRaised( display, w );
}

/**
 * Close a window
 *
 * @param w - id of window to close
 */
void WindowManager::closeWindow( Window w ) const
{
	sendClientMessage( w, "_NET_CLOSE_WINDOW" );
}

/**
 * Return currently active window
 */
Window WindowManager::getActiveWindow() const
{
	WindowManager::Property <Window> p;

	if( !p.getProperty( display, DefaultRootWindow( display ), XA_WINDOW, 
		"_NET_ACTIVE_WINDOW" ) )
		return 0;

	return *p.getData();
}

/**
 * Return first window whose title matches pattern
 *
 * @param pattern - a pattern with wildcards (*?)
 */
Window WindowManager::getWindowFromName( string pattern ) const
{
	WindowManager::WindowList list = getClientList();

	if( !list.count() )
		return 0;

	Window w;

	for( ; (w = list.fetch()); )
		if( isValid( w ) &&
			matchesPattern( getWindowTitle( w ).c_str(), pattern.c_str() ) )
			return w;

	return 0;	
}

/**
 * Return title of a window
 *
 * @param w - window in question
 */
string WindowManager::getWindowTitle( Window w ) const
{
	WindowManager::Property <char> p;
	
	if( !p.getProperty( display, w, XA_STRING, "WM_NAME" ) &&
		!p.getProperty( display, w, 
			XInternAtom( display, "UTF8_STRING", False ), 
			"_NET_WM_NAME" ) )
		return "";

	string s = p.getData();

	return s;
}

/**
 * Return list of all application windows
 */
WindowManager::WindowList WindowManager::getClientList() const
{
	WindowManager::Property <Window> *p = 
		new WindowManager::Property <Window>();

	if( !p->getProperty( display, root, XA_WINDOW, "_NET_CLIENT_LIST" ) )
		p->getProperty( display, root, XA_CARDINAL, "_WIN_CLIENT_LIST" );

	return WindowManager::WindowList( p );
}

/**
 * Send a ClientMessage to some window
 *
 * @param w - target window
 * @param msg - message to send
 * @param data0 - data (optional)
 * @param data1 - data (optional)
 * @param data2 - data (optional)
 * @param data3 - data (optional)
 * @param data4 - data (optional)
 */
void WindowManager::sendClientMessage( Window w, const char *msg, 
	unsigned long data0, unsigned long data1, unsigned long data2, 
	unsigned long data3, unsigned long data4 ) const
{
	XEvent event;
	long mask = SubstructureRedirectMask | SubstructureNotifyMask;

	event.xclient.type = ClientMessage;
	event.xclient.serial = 0;
	event.xclient.send_event = True;
	event.xclient.message_type = XInternAtom( display, msg, False );
	event.xclient.window = w;
	event.xclient.format = 32;
	event.xclient.data.l[0] = data0;
	event.xclient.data.l[1] = data1;
	event.xclient.data.l[2] = data2;
	event.xclient.data.l[3] = data3;
	event.xclient.data.l[4] = data4;

	XSendEvent( display, root, False, mask, &event );
}

/**
 * Intercept X errors
 */
int WindowManager::throwError( Display *, XErrorEvent * )
{
	return False;
}

/**
 * Returns if pattern matches string
 *
 * @param literal - string to match
 * @param pattern - pattern with optional wildcard characters (*?)
 */
const bool WindowManager::matchesPattern( const char *literal, 
	const char *pattern ) const
{
	for( ; *pattern; )
	{
		switch( *pattern )
		{
			case '*':
				{
					const char *match = pattern;

					while( *match == '*' || *match == '?' )
						match++;

					if( !*match )
						return true;

					const char *wildcard;
					char save;
					size_t length;

					if( (length = strcspn( match, "*?" )) )
					{
						wildcard = match+length;
						save = *wildcard;
						*((char *) wildcard) = 0;
					}

					const char *last = NULL;
					const char *pos;

					for( pos = literal;
						(pos = strstr( pos, match ));
						pos++ )
						last = pos;

					if( wildcard )
						*((char *) wildcard) = save;

					if( !last )
						return false;

					literal = last+length;

					if( !wildcard )
					{
						if( *literal )
							return false;

						return true;
					}

					pattern = wildcard;
				}
				break;
			case '?':
				if( !*(literal++) )
					return false;
				pattern++;
				break;
			default:
				if( !*literal ||
					*pattern != *literal )
					return false;
				literal++;
				pattern++;
				break;
		}
	}

	if( *literal )
		return false;

	return true;
}
